import os
import json
from enum import Enum
from typing import Dict, Any, List, Optional, Union, Callable
from pathlib import Path
from functools import partial
from dataclasses import dataclass, field
from deprecation import deprecated

import hashlib

from fal.dbt.integration.parse import normalize_path
from fal.dbt.integration.project import DbtModel, FalDbt, FAL

from dbt.contracts.results import RunStatus
from dbt.config.runtime import RuntimeConfig
from fal.dbt.integration.logger import LOGGER

from fal.dbt.telemetry import telemetry

from dbt.contracts.graph.nodes import ColumnInfo


class TimingType(Enum):
    PRE = "pre"
    POST = "post"

    def for_script(self):
        if self == TimingType.PRE:
            return "before"
        elif self == TimingType.POST:
            return "after"
        else:
            raise ValueError(f"Unknown timing type: {self}")

    def for_hook(self):
        return self.value

    def __str__(self):
        return self.for_hook()


class Hook:
    path: str
    arguments: Dict[str, Any]


@dataclass
class LocalHook(Hook):
    path: str
    arguments: Dict[str, Any] = field(default_factory=dict)


@dataclass
class IsolatedHook(Hook):
    path: str
    environment_name: str
    arguments: Dict[str, Any] = field(default_factory=dict)


def _is_local_environment(environment_name: str) -> None:
    return environment_name == "local"


def create_hook(raw_hook: Any, default_environment_name: Optional[str] = None) -> Hook:
    if isinstance(raw_hook, str):
        raw_hook = {"path": raw_hook}

    if not isinstance(raw_hook, dict):
        raise ValueError(f"Unrecognized hook value: {raw_hook}")

    if "path" not in raw_hook:
        raise ValueError(f"A hook must specify path.")

    environment_name = raw_hook.get("environment", default_environment_name)
    if environment_name and not _is_local_environment(environment_name):
        return IsolatedHook(
            raw_hook["path"],
            environment_name,
            raw_hook.get("with", {}),
        )
    else:
        return LocalHook(raw_hook["path"], raw_hook.get("with", {}))


@dataclass
class CurrentAdapterResponse:
    message: str
    code: Optional[str]
    rows_affected: Optional[int]


@dataclass
class CurrentModel:
    name: str
    alias: str
    status: RunStatus
    columns: Dict[str, ColumnInfo]
    tests: List[Any]
    meta: Dict[Any, Any]
    is_incremental: str
    adapter_response: Optional[CurrentAdapterResponse]


@dataclass
class CurrentTest:
    name: str
    model_name: str
    column: str
    status: str

    @property
    @deprecated(details="Use 'model_name' instead")
    def modelname(self):
        return self.model_name


@dataclass
class ContextConfig:
    target_path: Path

    def __init__(self, config: RuntimeConfig):
        self.target_path = Path(
            os.path.realpath(os.path.join(config.project_root, config.target_path))
        )


@dataclass
class ContextTarget:
    def __init__(self, config: RuntimeConfig):
        self.profile_name = config.profile_name
        self.name = config.target_name
        self.threads = config.threads
        self.type = config.credentials.type
        self.database = config.credentials.database
        self.schema = config.credentials.schema


@dataclass
class Context:
    current_model: Union[CurrentModel, None]
    config: ContextConfig
    target: ContextTarget
    _arguments: Optional[Dict[str, Any]] = field(repr=False, default=None)

    @property
    def arguments(self) -> Dict[str, Any]:
        if self._arguments is None:
            raise ValueError(
                "'context.arguments' is only accessible from hooks, "
                "not from scripts/models"
            )
        return self._arguments


@dataclass(frozen=True, init=False)
class FalScript:
    model: Optional[DbtModel]
    path: Path
    faldbt: FalDbt
    hook_arguments: Optional[Dict[str, Any]]
    is_hook: bool
    timing_type: Optional[TimingType]

    def __init__(
        self,
        faldbt: FalDbt,
        model: Optional[DbtModel],
        path: Union[str, Path],
        hook_arguments: Optional[Dict[str, Any]] = None,
        is_hook: bool = False,
        is_model: bool = False,
        timing_type: Optional[TimingType] = None,
    ):
        # Necessary because of frozen=True
        object.__setattr__(self, "model", model)
        object.__setattr__(
            self, "path", path if is_model else normalize_path(faldbt.scripts_dir, path)
        )
        object.__setattr__(self, "faldbt", faldbt)
        object.__setattr__(self, "hook_arguments", hook_arguments)
        object.__setattr__(self, "is_hook", is_hook)
        object.__setattr__(self, "timing_type", timing_type)

        self._telemetry()

    def _telemetry(self):
        try:
            _is_global = self.model is None
            _is_hook = self.is_hook
            _timing_type = str(self.timing_type) if self.timing_type else None
            _is_model = self.model and self.model.python_model == self.path
            _path_hash = hashlib.md5(str(self.path).encode()).hexdigest()

            _script_timing_desc = (
                self.timing_type.for_script() if self.timing_type else "error"
            )
            _hook_timing_desc = (
                self.timing_type.for_hook() if self.timing_type else "error"
            )

            if _is_global:
                _script_desc = f"{_script_timing_desc}-global"
                # globals are not hooks nor scripts
                _is_hook = None

            elif _is_model:
                _script_desc = "fal-model"
                _is_global = None
                # models are not hooks nor scripts
                _is_hook = None

            else:
                if _is_hook:
                    _script_desc = f"{_hook_timing_desc}-hook"
                else:
                    _script_desc = f"{_script_timing_desc}-script"

            telemetry.log_api(
                action="falscript_initialized",
                additional_props={
                    "is_global": _is_global,
                    "is_hook": _is_hook,
                    "is_model": _is_model,
                    "script_timing_type": _timing_type,
                    "script_desc": _script_desc,
                    "script_path": _path_hash,
                },
            )
        except:
            # Ignore telemetry errors
            pass

    @classmethod
    def from_hook(
        cls, faldbt: FalDbt, model: DbtModel, hook: Hook, timing_type: TimingType
    ):
        """
        Creates a FalScript from a hook
        """
        assert isinstance(hook, LocalHook)
        return cls(
            faldbt=faldbt,
            model=model,
            path=hook.path,
            hook_arguments=hook.arguments,
            is_hook=True,
            timing_type=timing_type,
        )

    @classmethod
    def model_script(cls, faldbt: FalDbt, model: DbtModel):
        assert model.python_model, "path for Python models must be set"
        return FalScript(
            faldbt=faldbt, model=model, path=model.python_model, is_model=True
        )

    def exec(self):
        """
        Executes the script
        """
        # Enable local imports
        try:
            source_code = python_from_file(self.path)
            program = compile(source_code, self.path, "exec")

            exec_globals = {
                "__name__": "__main__",
                "context": self._build_script_context(),
                "ref": self.faldbt.ref,
                "source": self.faldbt.source,
                "list_models": self.faldbt.list_models,
                "list_models_ids": self.faldbt.list_models_ids,
                "list_sources": self.faldbt.list_sources,
                "list_features": self.faldbt.list_features,
                "execute_sql": self.faldbt.execute_sql,
            }

            if not self.is_hook:
                exec_globals["write_to_source"] = self.faldbt.write_to_source

                if self.model is not None:
                    # Hard-wire the model
                    exec_globals["write_to_model"] = partial(
                        self.faldbt.write_to_model,
                        target_1=self.model.name,
                        target_2=None,
                    )

            else:
                exec_globals["write_to_source"] = _not_allowed_function_maker(
                    "write_to_source"
                )
                exec_globals["write_to_model"] = _not_allowed_function_maker(
                    "write_to_model"
                )
            exec(program, exec_globals)
        finally:
            pass

    @property
    def relative_path(self):
        if self.is_model:
            return self.path.relative_to(self.faldbt.project_dir)
        else:
            return self.path.relative_to(self.faldbt.scripts_dir)

    @property
    def id(self):
        if self.is_model:
            return f"(model: {self.relative_path})"
        else:
            return f"({self.model_name}, {self.relative_path})"

    @property
    def is_global(self):
        return self.model is None

    @property
    def is_model(self):
        if self.model is not None and self.model.python_model is not None:
            return self.model.python_model == self.path

    @property
    def model_name(self):
        return "<GLOBAL>" if self.is_global else self.model.name  # type: ignore

    def _build_script_context(self) -> Context:
        config: RuntimeConfig = self.faldbt._config
        context_config = ContextConfig(config)
        target = ContextTarget(config)

        if self.is_global:
            return Context(current_model=None, target=target, config=context_config)

        model: DbtModel = self.model  # type: ignore

        meta = model.meta or {}
        _del_key(meta, FAL)

        tests = _process_tests(model.tests)

        current_adapter_response = None
        if model.adapter_response:
            current_adapter_response = CurrentAdapterResponse(
                message=str(model.adapter_response),
                code=model.adapter_response.code,
                rows_affected=model.adapter_response.rows_affected,
            )

        current_model = CurrentModel(
            name=model.name,
            alias=model.alias,
            status=model.status,
            columns=model.columns,
            tests=tests,
            meta=meta,
            is_incremental=model.is_incremental,
            adapter_response=current_adapter_response,
        )

        return Context(
            current_model=current_model,
            target=target,
            config=context_config,
            _arguments=self.hook_arguments,
        )


def _del_key(dict: Dict[str, Any], key: str):
    try:
        del dict[key]
    except KeyError:
        pass


def _process_tests(tests: List[Any]):
    return list(
        map(
            lambda test: CurrentTest(
                name=test.name,
                column=test.column,
                status=test.status,
                model_name=test.model,
            ),
            tests,
        )
    )


def python_from_file(path: Path) -> str:
    with open(path) as file:
        raw_source_code = file.read()
        if path.suffix == ".ipynb":
            raw_source_code = _process_ipynb(raw_source_code)
    return raw_source_code


def _process_ipynb(raw_source_code: str) -> str:
    def strip_magic(source: List[str]) -> List[str]:
        NOTEBOOK_LIB = "faldbt.magics"
        return [item for item in source if item[0] != "%" and NOTEBOOK_LIB not in item]

    ipynb_struct = json.loads(raw_source_code)

    script_list = []
    for cell in ipynb_struct["cells"]:
        if cell["cell_type"] == "code":
            source = strip_magic(cell["source"])
            script_list.append("".join(source))

    joined_script = "\n #cell \n".join(script_list)

    LOGGER.debug(f"Joined .ipynb cells to:\n{joined_script}")

    return joined_script


def _not_allowed_function_maker(function_name: str) -> Callable[[Any], None]:
    def not_allowed_function(*args, **kwargs):
        raise Exception(
            (
                f"{function_name} is not allowed in hooks."
                " Consider using a Python model."
            )
        )

    return not_allowed_function



from __future__ import annotations

from dataclasses import dataclass, replace
from pathlib import Path
from typing import Any, ClassVar, Dict, Iterator, List, Optional, Tuple
import importlib_metadata

from dbt.events.adapter_endpoint import AdapterLogger
from dbt.exceptions import DbtRuntimeError
from dbt.config.runtime import RuntimeConfig

from isolate.backends import BaseEnvironment, BasicCallable, EnvironmentConnection
from fal_serverless import FalServerlessKeyCredentials, LocalHost
from fal_serverless.api import Host, FalServerlessHost

from . import cache_static
from .yaml_helper import load_yaml


CONFIG_KEYS_TO_IGNORE = ["host", "remote_type", "type", "name", "machine_type"]
REMOTE_TYPES_DICT = {
    "venv": "virtualenv",
    "conda": "conda",
}

logger = AdapterLogger("fal")


class FalParseError(Exception):
    pass


@dataclass
class LocalConnection(EnvironmentConnection):
    def run(self, executable: BasicCallable, *args, **kwargs) -> Any:
        return executable(*args, **kwargs)


@dataclass
class EnvironmentDefinition():
    host: Host
    kind: str
    config: dict[Any, Any]
    machine_type: str = "S"


def fetch_environment(
    project_root: str,
    environment_name: str,
    machine_type: str = "S",
    credentials: Optional[Any] = None,
) -> Tuple[EnvironmentDefinition, bool]:
    """Fetch the environment with the given name from the project's
    fal_project.yml file."""
    # Local is a special environment where it doesn't need to be defined
    # since it will mirror user's execution context directly.
    if environment_name == "local":
        if credentials.host:
            logger.warning(
                "`local` environments will be executed on fal cloud."
                + "If you don't want to use fal cloud, you can change your "
                + "profile target to one where fal doesn't have credentials"
            )
            host = FalServerlessHost(
                url=credentials.host,
                credentials=FalServerlessKeyCredentials(credentials.key_id, credentials.key_secret))
            return EnvironmentDefinition(
                host=host,
                kind="virtualenv",
                machine_type=machine_type,
                config={"name": "", "type": "venv"}), False

        return EnvironmentDefinition(host=LocalHost(), kind="local", config={}), True

    try:
        environments = load_environments(project_root, machine_type, credentials)
    except Exception as exc:
        raise DbtRuntimeError(
            "Error loading environments from fal_project.yml"
        ) from exc

    if environment_name not in environments:
        raise DbtRuntimeError(
            f"Environment '{environment_name}' was used but not defined in fal_project.yml"
        )

    return environments[environment_name], False


def db_adapter_config(config: RuntimeConfig) -> RuntimeConfig:
    """Return a config object that has the database adapter as its primary. Only
    applicable when the underlying db adapter is encapsulated."""
    if hasattr(config, "sql_adapter_credentials"):
        new_config = replace(config, credentials=config.sql_adapter_credentials)
        new_config.python_adapter_credentials = config.credentials
    else:
        new_config = config

    return new_config


def load_environments(
    base_dir: str, machine_type: str = "S", credentials: Optional[Any] = None
) -> Dict[str, EnvironmentDefinition]:
    import os
    fal_project_path = os.path.join(base_dir, "fal_project.yml")
    if not os.path.exists(fal_project_path):
        raise FalParseError(f"{fal_project_path} must exist to define environments")

    fal_project = load_yaml(fal_project_path)
    environments = {}
    for environment in fal_project.get("environments", []):
        env_name = _get_required_key(environment, "name")
        if _is_local_environment(env_name):
            raise FalParseError(
                f"Environment name conflicts with a reserved name: {env_name}."
            )

        env_kind = _get_required_key(environment, "type")
        if environments.get(env_name) is not None:
            raise FalParseError("Environment names must be unique.")

        environments[env_name] = create_environment(
            env_name, env_kind, environment, machine_type, credentials
        )

    return environments


def create_environment(
    name: str,
    kind: str,
    config: Dict[str, Any],
    machine_type: str = "S",
    credentials: Optional[Any] = None,
) -> EnvironmentDefinition:
    if kind not in ["venv", "conda"]:
        raise ValueError(
            f"Invalid environment type (of {kind}) for {name}. Please choose from: "
            + "venv, conda."
        )

    kind = kind if kind == "conda" else "virtualenv"

    parsed_config = {
        key: val for key, val in config.items() if key not in CONFIG_KEYS_TO_IGNORE
    }

    if credentials.key_secret and credentials.key_id:
        host = FalServerlessHost(
            url=credentials.host,
            credentials=FalServerlessKeyCredentials(credentials.key_id, credentials.key_secret))
    else:
        host = LocalHost()
    return EnvironmentDefinition(
        host=host,
        kind=kind,
        config=parsed_config,
        machine_type=machine_type)


def _is_local_environment(environment_name: str) -> bool:
    return environment_name == "local"


def _get_required_key(data: Dict[str, Any], name: str) -> Any:
    if name not in data:
        raise FalParseError("Missing required key: " + name)
    return data[name]

def _parse_remote_config(
    config: Dict[str, Any], parsed_config: Dict[str, Any]
) -> Dict[str, Any]:
    assert config.get("remote_type"), "remote_type needs to be specified."

    remote_type = REMOTE_TYPES_DICT.get(config["remote_type"])

    assert (
        remote_type
    ), f"{config['remote_type']} not recognised. Available remote types: {list(REMOTE_TYPES_DICT.keys())}"

    env_definition = {
        "kind": remote_type,
        "configuration": parsed_config,
    }

    return {
        "host": config.get("host"),
        "target_environments": [env_definition],
    }

def _get_package_from_type(adapter_type: str):
    SPECIAL_ADAPTERS = {
        # Documented in https://docs.getdbt.com/docs/supported-data-platforms#community-adapters
        "athena": "dbt-athena-community",
    }
    return SPECIAL_ADAPTERS.get(adapter_type, f"dbt-{adapter_type}")

def _get_dbt_packages(
    adapter_type: str,
    is_teleport: bool = False,
    is_remote: bool = False,
) -> Iterator[Tuple[str, Optional[str]]]:
    dbt_adapter = _get_package_from_type(adapter_type)
    for dbt_plugin_name in [dbt_adapter]:
        distribution = importlib_metadata.distribution(dbt_plugin_name)

        yield dbt_plugin_name, distribution.version

    try:
        dbt_fal_version = importlib_metadata.version("dbt-fal")
    except importlib_metadata.PackageNotFoundError:
        # It might not be installed.
        return None

    dbt_fal_dep = "dbt-fal"
    dbt_fal_extras = _find_adapter_extras(dbt_fal_dep, dbt_adapter)
    if is_teleport:
        dbt_fal_extras.add("teleport")
    dbt_fal_suffix = ""

    if _version_is_prerelease(dbt_fal_version):
        if is_remote:
            # If it's a pre-release and it's remote, its likely us developing, so we try installing
            # from Github and we can get the custom branch name from FAL_GITHUB_BRANCH environment variable
            # TODO: Handle pre-release on PyPI. How should we approach that?
            import os

            branch_name = os.environ.get("FAL_GITHUB_BRANCH", "main")

            dbt_fal_suffix = f" @ git+https://github.com/fal-ai/fal.git@{branch_name}#subdirectory=projects/adapter"
            dbt_fal_version = None
        else:
            dbt_fal_path = _get_project_root_path("adapter")
            if dbt_fal_path is not None:
                # Can be a pre-release from PyPI
                dbt_fal_dep = str(dbt_fal_path)
                dbt_fal_version = None

    dbt_fal = f"{dbt_fal_dep}[{' ,'.join(dbt_fal_extras)}]{dbt_fal_suffix}"
    yield dbt_fal, dbt_fal_version


def _find_adapter_extras(package: str, plugin_package: str) -> set[str]:
    import pkgutil
    import dbt.adapters

    all_extras = _get_extras(package)
    available_plugins = {
        module_info.name
        for module_info in pkgutil.iter_modules(dbt.adapters.__path__)
        if module_info.ispkg and module_info.name in plugin_package
    }
    return available_plugins.intersection(all_extras)


def _get_extras(package: str) -> list[str]:
    import importlib_metadata

    dist = importlib_metadata.distribution(package)
    return dist.metadata.get_all("Provides-Extra", [])


def _version_is_prerelease(raw_version: str) -> bool:
    from packaging.version import Version

    package_version = Version(raw_version)
    return package_version.is_prerelease


def _get_project_root_path(package: str) -> Path:
    from dbt.adapters import fal

    # If this is a development version, we'll install
    # the current fal itself.
    path = Path(fal.__file__)
    while path is not None:
        if (path.parent / ".git").exists():
            break
        path = path.parent
    return path / package


def get_default_requirements(
    adapter_type: str,
    is_teleport: bool = False,
    is_remote: bool = False,
) -> Iterator[Tuple[str, Optional[str]]]:
    yield from _get_dbt_packages(adapter_type, is_teleport, is_remote)

@cache_static
def get_default_pip_dependencies(
    adapter_type: str,
    is_teleport: bool = False,
    is_remote: bool = False,
) -> List[str]:
    return [
        f"{package}=={version}" if version else package
        for package, version in get_default_requirements(
            adapter_type, is_teleport, is_remote
        )
    ]



from pathlib import Path
from typing import List, Optional, Dict, Any

from fal.dbt.packages.environments.base import BASE_CACHE_DIR, BaseEnvironment
from fal.dbt.packages.environments.conda import CondaEnvironment
from fal.dbt.packages.environments.virtual_env import VirtualPythonEnvironment

REGISTERED_ENVIRONMENTS: Dict[str, BaseEnvironment] = {
    "conda": CondaEnvironment,
    "venv": VirtualPythonEnvironment,
}


def create_environment(name: str, kind: str, config: Dict[str, Any]) -> BaseEnvironment:
    env_type = REGISTERED_ENVIRONMENTS.get(kind)
    if env_type is None:
        raise ValueError(
            f"Invalid environment type (of {kind}) for {name}. Please choose from: "
            + ", ".join(REGISTERED_ENVIRONMENTS.keys())
        )

    return env_type.from_config(config)



import os
from dataclasses import dataclass
import glob
from pathlib import Path
from typing import Any, List, Dict, Optional, Union, TYPE_CHECKING

from dbt.contracts.project import Project as ProjectContract
from dbt.config import RuntimeConfig, Project
from dbt.config.utils import parse_cli_vars as dbt_parse_cli_vars
from dbt.contracts.graph.manifest import Manifest
from dbt.contracts.results import RunResultsArtifact, FreshnessExecutionResultArtifact
from dbt.contracts.project import UserConfig
from dbt.config.profile import read_user_config

from dbt.exceptions import IncompatibleSchemaError, DbtRuntimeError

from fal.dbt.utils import cache_static

from fal.dbt.integration.logger import LOGGER
from fal.dbt.integration.utils.yaml_helper import load_yaml
from fal.dbt.telemetry import telemetry

if TYPE_CHECKING:
    from fal.dbt.packages.environments import BaseEnvironment

FAL_SCRIPTS_PATH = "fal-scripts-path"
FAL_MODELS_PATHS = "fal-models-paths"


class FalParseError(Exception):
    pass


def get_dbt_user_config(profiles_dir: str) -> UserConfig:
    return read_user_config(profiles_dir)


@dataclass
class RuntimeArgs:
    project_dir: str
    profiles_dir: str
    threads: Optional[int]
    single_threaded: bool
    profile: Optional[str]
    target: Optional[str]
    vars: Dict[str, str]


def load_dbt_project_contract(project_dir: str) -> ProjectContract:
    partial_project = Project.partial_load(project_dir)
    contract = ProjectContract.from_dict(partial_project.project_dict)
    if not hasattr(contract, "model_paths") or contract.model_paths is None:
        setattr(contract, "model_paths", contract.source_paths)
    if not hasattr(contract, "seed_paths") or contract.seed_paths is None:
        setattr(contract, "seed_paths", contract.data_paths)
    return contract


def get_dbt_config(
    *,
    project_dir: str,
    profiles_dir: str,
    profile_target: Optional[str] = None,
    threads: Optional[int] = None,
    profile: Optional[str] = None,
    args_vars: str = "{}",
) -> RuntimeConfig:
    # Construct a phony config
    import os

    vars = get_vars_dict(project_dir, args_vars)
    args = RuntimeArgs(
        project_dir=project_dir,
        profiles_dir=profiles_dir,
        threads=threads,
        single_threaded=False,
        profile=profile,
        target=profile_target,
        vars=vars,
    )

    if project_dir and not "PYTEST_CURRENT_TEST" in os.environ:
        # HACK: initializing dbt-fal requires cwd to be project_dir
        # TODO: this doesn't work in pytest + Github Actions
        owd = os.getcwd()
        os.chdir(project_dir)
        config = RuntimeConfig.from_args(args)
        os.chdir(owd)
    else:
        config = RuntimeConfig.from_args(args)

    # HACK: issue in dbt-core 1.5.0 https://github.com/dbt-labs/dbt-core/issues/7465
    env_target_path = os.getenv("DBT_TARGET_PATH")
    if env_target_path:
        config.target_path = env_target_path
    # TODO: should we check flags too?

    return config


def get_vars_dict(project_dir: str, args_vars: str) -> Dict[str, Any]:
    project_contract = load_dbt_project_contract(project_dir)

    # NOTE: This happens usually inside unit tests
    vars = (project_contract is not None and project_contract.vars) or {}
    cli_vars = parse_cli_vars(args_vars)

    # cli_vars have higher priority
    return {**vars, **cli_vars}

def parse_cli_vars(args_vars: str) -> Dict[str, Any]:
    if not args_vars:
        return {}

    try:
        return dbt_parse_cli_vars(args_vars)
    except DbtRuntimeError as exc:
        raise FalParseError(exc)


@cache_static
def get_fal_models_dirs(project_dir: str, args_vars: str) -> List[str]:
    vars = get_vars_dict(project_dir, args_vars)
    model_paths = vars.get(FAL_MODELS_PATHS) or []
    if not model_paths:
        # None or empty list
        LOGGER.warn(
            f"Variable '{FAL_MODELS_PATHS}' not defined. Locate fal-format "
            "Python models in a separate model directory and set it as the variable. "
            "e.g. {FAL_MODELS_PATHS}: ['fal_models']"
        )

        telemetry.log_api(action="fal_models_paths_not_set")

    if not isinstance(model_paths, list):
        raise FalParseError(
            f"Error parsing '{FAL_MODELS_PATHS}'. Expected list of strings and got '{type(model_paths)}'"
        )

    return model_paths


def get_scripts_dir(project_dir: str, args_vars: str) -> str:
    vars = get_vars_dict(project_dir, args_vars)
    scripts_dir = vars.get(FAL_SCRIPTS_PATH, project_dir)

    if not isinstance(scripts_dir, str):
        raise FalParseError(
            f"Error parsing '{FAL_SCRIPTS_PATH}'. Expected string and got '{type(scripts_dir)}'"
        )

    return os.path.join(project_dir, scripts_dir)


def get_dbt_manifest(config) -> Manifest:
    from dbt.parser.manifest import ManifestLoader

    return ManifestLoader.get_full_manifest(config)


def get_dbt_sources_artifact(project_dir: str, config: RuntimeConfig):
    # Changed in dbt 1.5.0 to use path relative to CWD instead of path relative to project_dir
    sources_path = os.path.join(config.target_path, "sources.json")
    try:
        return FreshnessExecutionResultArtifact.read_and_check_versions(sources_path)

    except IncompatibleSchemaError as exc:
        # TODO: add test for this case
        exc.add_filename(sources_path)
        raise
    except DbtRuntimeError as exc:
        LOGGER.warn("Could not read dbt sources artifact")
        return None


def get_dbt_results(project_dir: str, config: RuntimeConfig) -> Optional[RunResultsArtifact]:
    # Changed in dbt 1.5.0 to use path relative to CWD instead of path relative to project_dir
    results_path = os.path.join(config.target_path, "run_results.json")
    try:
        return RunResultsArtifact.read_and_check_versions(results_path)

    except IncompatibleSchemaError as exc:
        # TODO: add test for this case
        exc.add_filename(results_path)
        raise
    except DbtRuntimeError as exc:
        LOGGER.warn("Could not read dbt run_results artifact")
        return None


def get_scripts_list(scripts_dir: str) -> List[str]:
    scripts_path = Path(scripts_dir)
    return list(map(str, [*scripts_path.rglob("*.py"), *scripts_path.rglob("*.ipynb")]))


def get_global_script_configs(source_dirs: List[Path]) -> Dict[str, List[str]]:
    global_scripts = {"before": [], "after": []}
    for source_dir in source_dirs:
        # Scan directories for .yml files
        schema_files = glob.glob(os.path.join(source_dir, "**.yml"), recursive=True)
        # Scan directories for .yaml files
        schema_files += glob.glob(os.path.join(source_dir, "**.yaml"), recursive=True)
        for file in schema_files:
            schema_yml = load_yaml(file)
            if schema_yml is not None:
                fal_config = schema_yml.get("fal", None)
                if fal_config is not None:
                    # sometimes `scripts` can *be* there and still be None
                    script_paths = fal_config.get("scripts") or []
                    if isinstance(script_paths, list):
                        global_scripts["after"] += script_paths
                    else:
                        global_scripts["before"] += script_paths.get("before") or []
                        global_scripts["after"] += script_paths.get("after") or []
            else:
                raise FalParseError("Error parsing the schema file " + file)

    return global_scripts


def _get_required_key(data: Dict[str, Any], name: str) -> Any:
    if name not in data:
        raise FalParseError("Missing required key: " + name)
    return data[name]


def load_environments(base_dir: str) -> Dict[str, "BaseEnvironment"]:
    from fal.dbt.packages.environments import create_environment
    from fal.dbt.fal_script import _is_local_environment

    try:
        fal_project_path = os.path.join(base_dir, "fal_project.yml")
        if not os.path.exists(fal_project_path):
            raise FalParseError(f"{fal_project_path} must exist to define environments")

        fal_project = load_yaml(fal_project_path)

        environments = {}
        for environment in fal_project.get("environments", []):
            env_name = _get_required_key(environment, "name")
            if _is_local_environment(env_name):
                raise FalParseError(
                    f"Environment name conflicts with a reserved name: {env_name}."
                )

            env_kind = _get_required_key(environment, "type")
            environments[env_name] = create_environment(env_name, env_kind, environment)
        return environments
    except FalParseError as e:
        raise RuntimeError("Error loading environments from fal_project.yml") from e


def normalize_path(base: str, path: Union[Path, str]):
    real_base = os.path.realpath(os.path.normpath(base))
    return Path(os.path.realpath(os.path.join(real_base, path)))


def normalize_paths(
    base: str, paths: Union[List[Path], List[str], List[Union[Path, str]]]
):
    return list(map(lambda path: normalize_path(base, path), paths))



import itertools
import re
from dataclasses import dataclass
from typing import List, Optional, Union, Iterator
from fal.dbt.node_graph import NodeGraph
from fal.dbt.integration.project import CompileArgs, FalDbt, FalGeneralException
from dbt.task.compile import CompileTask
from enum import Enum
from functools import reduce
import networkx as nx


class ExecutionPlan:
    """
    Represents a fal flow excution
    """

    before_scripts: List[str]
    dbt_models: List[str]
    after_scripts: List[str]
    project_name: str

    def __init__(self, unique_ids: List[str], project_name):
        self.before_scripts = []
        self.dbt_models = []
        self.after_scripts = []
        self.project_name = project_name
        for id in unique_ids:
            if _is_before_script(id):
                self.before_scripts.append(id)
            elif _is_after_script(id):
                self.after_scripts.append(id)
            else:
                self.dbt_models.append(id)

    @property
    def nodes(self) -> List[str]:
        return self.before_scripts + self.after_scripts + self.dbt_models

    @classmethod
    def create_plan_from_graph(cls, parsed, nodeGraph: NodeGraph, fal_dbt: FalDbt):
        """
        Creates and ExecutionPlan from the cli arguments
        """
        unique_ids = list(nodeGraph.graph.nodes.keys())

        ids_to_execute = unique_ids

        if parsed.select:
            ids_to_execute = _filter_node_ids(
                unique_ids, fal_dbt, list(parsed.select), nodeGraph
            )

        ids_to_exclude = []
        if "exclude" in parsed and parsed.exclude:
            ids_to_exclude = _filter_node_ids(
                unique_ids, fal_dbt, list(parsed.exclude), nodeGraph
            )

        ids_to_execute = [i for i in ids_to_execute if i not in ids_to_exclude]

        # Remove non-model nodes (sources, maybe more?) by making sure they are in the node_lookup dict
        ids_to_execute = [i for i in ids_to_execute if i in nodeGraph.node_lookup]

        return cls(list(set(ids_to_execute)), fal_dbt.project_name)


@dataclass
class SelectionUnion:
    components: List[str]


@dataclass
class SelectionIntersection:
    components: List[str]


def parse_union(
    components: List[str],
) -> SelectionUnion:
    # Based on the original implemention at dbt-core.

    # turn ['a b', 'c'] -> ['a', 'b', 'c']
    raw_specs = itertools.chain.from_iterable(r.split(OP_SET_UNION) for r in components)
    union_components = []

    # ['a', 'b', 'c,d'] -> union('a', 'b', intersection('c', 'd'))
    for raw_spec in raw_specs:
        union_components.append(
            SelectionIntersection(
                raw_spec.split(OP_SET_INTERSECTION),
            )
        )
    return SelectionUnion(
        union_components,
    )


def _filter_node_ids(
    unique_ids: List[str],
    fal_dbt: FalDbt,
    selectors: List[str],
    nodeGraph: NodeGraph,
) -> List[str]:
    """Filter list of unique_ids according to a selector."""
    output = set()

    union = parse_union(selectors)

    for intersection in union.components:
        try:
            plan_outputs = [
                set(SelectorPlan(selector, unique_ids, fal_dbt).execute(nodeGraph))
                for selector in intersection.components
                if selector
            ]
        except nx.NetworkXError:
            # When the user selects a non-existent node, don't fail immediately
            # but rather just continue processing the rest of selectors.
            plan_outputs = []

        if plan_outputs:
            output |= set.intersection(*plan_outputs)

    return list(output)


def _get_children_with_parents(node_id: str, nodeGraph: NodeGraph) -> List[str]:
    children = nodeGraph.get_descendants(node_id)
    output = reduce(lambda l, ch: l + nodeGraph.get_ancestors(ch), children, children)

    output = list(set(output))

    return output


def _expand_script(script_name: str, unique_ids: List[str]) -> List[str]:
    """
    Expands the selected script name to unique id format.
    for example [scripta.py] to [script.modelB.AFTER.scripta.py, script.modelA.BEFORE.scripta.py]
    """

    def contains_script_name(id: str):
        return script_name in id

    return list(filter(contains_script_name, unique_ids))


class SelectType(Enum):
    MODEL = 1
    SCRIPT = 2
    COMPLEX = 3
    TAG = 4


@dataclass(init=False)
class SelectorPlan:
    """
    Represents a single selector, for example in the command

    fal flow run --select script.py+

    script.py+ is the SelectorPlan with needs_children attribute set to true
    """

    unique_ids: List[str]
    children: bool
    children_levels: Optional[int]
    children_with_parents: bool
    parents: bool
    parents_levels: Optional[int]
    type: SelectType
    raw: str

    def __init__(self, selector: str, unique_ids: List[str], fal_dbt: FalDbt):
        self.raw = selector
        self.children_with_parents = OP_CHILDREN_WITH_PARENTS.match(selector)
        selector = OP_CHILDREN_WITH_PARENTS.rest(selector)

        self.parents = OP_PARENTS.match(selector)
        self.parents_levels = OP_PARENTS.depth(selector)
        selector = OP_PARENTS.rest(selector)

        self.children = OP_CHILDREN.match(selector)
        self.children_levels = OP_CHILDREN.depth(selector)
        selector = OP_CHILDREN.rest(selector)

        self.type = _to_select_type(selector)

        if self.type == SelectType.MODEL:
            self.unique_ids = [f"model.{fal_dbt.project_name}.{selector}"]
        elif self.type == SelectType.SCRIPT:
            self.unique_ids = _expand_script(selector, unique_ids)
        elif self.type == SelectType.TAG:
            self.unique_ids = unique_ids_from_tag_selector(selector, fal_dbt)
        elif self.type == SelectType.COMPLEX:
            self.unique_ids = unique_ids_from_complex_selector(selector, fal_dbt)

    def __post_init__(self):
        if self.children and self.children_with_parents:
            raise RuntimeError(
                f'Invalid node spec {self.raw} - "@" prefix and "+" suffix are incompatible'
            )

    def execute(self, nodeGraph: NodeGraph) -> Iterator[str]:
        for id in self.unique_ids:
            yield id

            if self.children:
                if self.children_levels is None:
                    children = nodeGraph.get_descendants(id)
                else:
                    children = nodeGraph.get_successors(id, self.children_levels)
                yield from children

            if self.parents:
                if self.parents_levels is None:
                    parents = nodeGraph.get_ancestors(id)
                else:
                    parents = nodeGraph.get_predecessors(id, self.parents_levels)
                yield from parents

            if self.children_with_parents:
                ids = _get_children_with_parents(id, nodeGraph)
                yield from ids


def unique_ids_from_complex_selector(select, fal_dbt: FalDbt) -> List[str]:
    args = CompileArgs(None, [select], [select], tuple(), fal_dbt._state, None)
    compile_task = CompileTask(args, fal_dbt._config, fal_dbt._manifest.native_manifest)
    compile_task._runtime_initialize()
    spec = compile_task.get_selection_spec()
    graph = compile_task.get_node_selector().get_graph_queue(spec)
    return list(graph.queued)

def unique_ids_from_tag_selector(selector: str, fal_dbt: FalDbt) -> List[str]:
    parts = selector.split(":")
    if len(parts) != 2:
        raise FalGeneralException(f"Expected selector tag:tag_value, got: {selector}")
    tag = parts[1]
    ids = [model.unique_id for model in fal_dbt.models if tag in model.node.tags]
    return ids


def _to_select_type(selector: str) -> SelectType:
    if ":" in selector:
        parts = selector.split(':')
        if parts[0] == "tag":
            return SelectType.TAG
        return SelectType.COMPLEX
    else:
        if _is_script_node(selector):
            return SelectType.SCRIPT
        else:
            return SelectType.MODEL


def _is_script_node(node_name: str) -> bool:
    return node_name.endswith(".py") or node_name.endswith(".ipynb")


class SelectorGraphOp:
    _regex: re.Pattern

    def __init__(self, regex: re.Pattern):
        self._regex = regex
        assert (
            "rest" in regex.groupindex
        ), 'rest must be in regex. Use `re.compile("something(?P<rest>.*)")`'

    def _select(self, selector: str, group: Union[str, int]) -> Optional[str]:
        match = self._regex.match(selector)
        if match:
            return match.group(group)

    def match(self, selector: str) -> bool:
        return self._select(selector, 0) is not None

    def rest(self, selector: str) -> str:
        rest = self._select(selector, "rest")
        if rest is not None:
            return rest
        return selector


class SelectorGraphOpDepth(SelectorGraphOp):
    def depth(self, selector: str) -> Optional[int]:
        depth = self._select(selector, "depth")
        if depth:
            return int(depth)


# Graph operators from their regex Patterns
OP_CHILDREN_WITH_PARENTS = SelectorGraphOp(re.compile("^\\@(?P<rest>.*)"))
OP_PARENTS = SelectorGraphOpDepth(re.compile("^(?P<depth>\\d*)\\+(?P<rest>.*)"))
OP_CHILDREN = SelectorGraphOpDepth(re.compile("(?P<rest>.*)\\+(?P<depth>\\d*)$"))

# Logic based set operators
OP_SET_UNION = " "
OP_SET_INTERSECTION = ","

IS_BEFORE_SCRIPT_REGEX = re.compile("^script\\..+\\.BEFORE\\..+\\.(py|ipynb)$")
IS_AFTER_SCRIPT_REGEX = re.compile("^script\\..+\\.AFTER\\..+\\.(py|ipynb)")


def _is_before_script(id: str) -> bool:
    return bool(IS_BEFORE_SCRIPT_REGEX.match(id))


def _is_after_script(id: str) -> bool:
    return bool(IS_AFTER_SCRIPT_REGEX.match(id))



from typing import Any
import six
from dbt.adapters.base.relation import BaseRelation
from dbt.contracts.connection import AdapterResponse
from dbt.adapters.fal.wrappers import FalCredentialsWrapper
import sqlalchemy
import pandas as pd
from dbt.adapters.base import BaseAdapter
from urllib.parse import quote_plus


def create_engine(adapter: BaseAdapter) -> Any:
    if isinstance(adapter.config.credentials, FalCredentialsWrapper):
        creds = adapter.config.credentials._db_creds
    else:
        # In isolated environment, credentials are AthenaCredentials
        creds = adapter.config.credentials
    conn_str = ("awsathena+rest://:@athena.{region_name}.amazonaws.com:443/"
                "{schema_name}?s3_staging_dir={s3_staging_dir}"
                "&location={location}&compression=snappy")
    return sqlalchemy.create_engine(conn_str.format(
        region_name=creds.region_name,
        schema_name=creds.schema,
        s3_staging_dir=quote_plus(creds.s3_staging_dir),
        location=quote_plus(creds.s3_staging_dir)))


def drop_relation_if_it_exists(adapter: BaseAdapter, relation: BaseRelation) -> None:
    if adapter.get_relation(
        database=relation.database,
        schema=relation.schema,
        identifier=relation.identifier,
    ):
        adapter.drop_relation(relation)


def write_df_to_relation(adapter, dataframe, relation, if_exists) -> AdapterResponse:

    assert adapter.type() == "athena"
    if isinstance(adapter.config.credentials, FalCredentialsWrapper):
        creds = adapter.config.credentials._db_creds
    else:
        # In isolated environment, credentials are AthenaCredentials
        creds = adapter.config.credentials

    # This is a quirk of dbt-athena-community, where they set
    # relation.schema = relation.identifier
    temp_relation = relation.replace_path(
        schema=relation.database,
        database=creds.database,
        # athena complanes when table location has x.__y
        identifier=f"dbt_fal_temp_{relation.schema}"
    )

    relation = temp_relation.replace_path(identifier=relation.schema)

    drop_relation_if_it_exists(adapter, temp_relation)

    alchemy_engine = create_engine(adapter)

    rows_affected = dataframe.to_sql(
        con=alchemy_engine,
        name=temp_relation.identifier,
        schema=temp_relation.schema,
        if_exists=if_exists,
        index=False,
    )

    adapter.cache.add(temp_relation)

    drop_relation_if_it_exists(adapter, relation)


    # athena doesn't let us rename relations, so we do it by hand
    stmt = f"create table {relation} as select * from {temp_relation} with data"
    adapter.execute(six.text_type(stmt).strip())
    adapter.cache.add(relation)
    adapter.drop_relation(temp_relation)

    adapter.commit_if_has_connection()
    return AdapterResponse("OK", rows_affected=rows_affected)

def read_relation_as_df(adapter: BaseAdapter, relation: BaseRelation) -> pd.DataFrame:
    alchemy_engine = create_engine(adapter)

    # This is dbt-athena-community quirk, table_name=relation.schema

    return pd.read_sql_table(
        con=alchemy_engine,
        table_name=relation.schema,
        schema=relation.database,
    )
